# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2018, 2019 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from __future__ import division

__all__ = ['']

import os
import marshal
from chemicals.utils import log, exp
from chemicals.utils import mixing_simple, none_and_length_check, Vm_to_rho
from fluids.constants import N_A, k
from thermo.utils import TDependentProperty, MixtureProperty
from thermo.chemical import ChemicalConstants


folder = os.path.join(os.path.dirname(__file__), 'Misc')





def loadChemicalConstants(data, rows=True):
    '''Accepts either a marshal-style list-of-lists with fixed indexes, or
    takes in the json-style dict-of-dicts-of-dicts.
    Returns a dictionary of ChemicalConstants indexed by their CASs.
    '''
    loaded_chemicals = {}
    # Question: What if every database is a per-datasource one
    # And I fit all methods to apolynom to within their range?
    # Then we have one additional database which holds the best data source.
    # That way, all coefficients are precisely sourced.

    def add_chemical(kwargs):
        # TODO: remove to skip a function call
        constants = ChemicalConstants(**kwargs)
        loaded_chemicals[constants.CAS] = constants


    if rows:
        for row in data:
            kwargs = dict(CAS=row[0], Tc=row[1], Pc=row[2], Vc=row[3], omega=row[4], Tb=row[5],
                         Tm=row[6], Tt=row[7], Pt=row[8], Hfus=row[9], Hsub=row[10], Hf=row[11],
                         dipole=row[12],
                         HeatCapacityGas=row[13], HeatCapacityLiquid=row[14],
                         HeatCapacitySolid=row[15],
                         ThermalConductivityLiquid=row[16], ThermalConductivityGas=row[17],
                         ViscosityLiquid=row[18], ViscosityGas=row[19],
                         EnthalpyVaporization=row[20], VaporPressure=row[21], VolumeLiquid=row[22],
                         SublimationPressure=row[23], EnthalpySublimation=row[24],
                         VolumeSolid=row[25], VolumeSupercriticalLiquid=row[26])
            add_chemical(kwargs)
    else:
        for CAS, item in data.items():
            kwargs= dict(CAS=CAS, Tc=item['Tc']['value'],
                    Pc=item['Pc']['value'],
                    Vc=item['Vc']['value'],
                    omega=item['omega']['value'],
                    Tb=item['Tb']['value'],
                    Tm=item['Tm']['value'],
                    Tt=item['Tt']['value'],
                    Pt=item['Pt']['value'],
                    Hfus=item['Hfus']['value'],
                    Hsub=item['Hsub']['value'],
                    Hf=item['Hf']['value'],
                    dipole=item['dipole']['value'])

            for prop_key, store in marshal_properties:
                try:
                    prop_data = item[prop_key]
                    Tmin, Tmax = prop_data['Tmin'], prop_data['Tmax']
                    coefficients = prop_data['coefficients']
                    if 'Tc' in prop_data:
                        kwargs[prop_key] = (Tmin, Tmax, prop_data['Tc'], coefficients)
                    else:
                        kwargs[prop_key] = (Tmin, Tmax, coefficients)
                except KeyError:
                    pass
#                    Tmin, Tmax, coefficients = None, None, None
#                kwargs[prop_key] = (Tmin, Tmax, coefficients)
            add_chemical(kwargs)
    return loaded_chemicals


def load_json_data(json_path):
    f = open(json_path, 'r')
    import json
    full_data = json.load(f)
    f.close()
    return full_data

def marshal_json_data(full_data, path):
    marshal_rows = []
    for CAS, data in full_data.items():
        row = [CAS]
        row.append(data['Tc']['value'])
        row.append(data['Pc']['value'])
        row.append(data['Vc']['value'])
        row.append(data['omega']['value'])
        row.append(data['Tb']['value'])
        row.append(data['Tm']['value'])
        row.append(data['Tt']['value'])
        row.append(data['Pt']['value'])
        row.append(data['Hfus']['value'])
        row.append(data['Hsub']['value'])
        row.append(data['Hf']['value'])
        row.append(data['dipole']['value'])

        for prop_key, store in marshal_properties:
            try:
                prop_data = data[prop_key]
                Tmin, Tmax = prop_data['Tmin'], prop_data['Tmax']
                coefficients = prop_data['coefficients']
                if 'Tc' in prop_data:
                    Tc = prop_data['Tc']
                    row = (Tmin, Tmax, Tc, coefficients)
                else:
                    row = (Tmin, Tmax, coefficients)
            except KeyError:
                row = (None, None, None)

            row.append(row)

        marshal_rows.append(row)

    f = open(path, 'wb')
    marshal.dump(marshal_rows, f, 2)
    f.close()
    return marshal_rows



marshal_properties = [('HeatCapacityGas', True),
           ('HeatCapacityLiquid', True),
           ('HeatCapacitySolid', True),

           ('ThermalConductivityLiquid', True),
           ('ThermalConductivityGas', True),

           ('ViscosityLiquid', True),
           ('ViscosityGas', True),

           ('EnthalpyVaporization', True),
           ('VaporPressure', True),
           ('VolumeLiquid', True),
           ('SublimationPressure', True),
           ('EnthalpySublimation', True),
           ('VolumeSolid', True),
           ('VolumeSupercriticalLiquid', True),
           ]


json_path = os.path.join(folder, 'constants dump.json')
binary_path = os.path.join(folder, 'binary dump.marshal')
skip = not os.path.exists(json_path)

loaded_chemicals = {}

if not skip:
    from_json = True
    if os.path.exists(binary_path):
        # get the changed dates for each file and only load from binary if
        # the binary file is newer
        json_mtime = os.path.getmtime(json_path)
        binary_mtime = os.path.getmtime(binary_path)
    
        if binary_mtime > json_mtime and os.path.getsize(binary_path) > 10000:
            from_json = False
    
    
    full_data = {}
    marshal_rows = []
    
    
    if from_json:
        full_data = load_json_data(json_path)
        loaded_chemicals = loadChemicalConstants(full_data, rows=False)
    
    
    marshal_data = from_json
    if marshal_data:
        try:
            marshal_rows = marshal_json_data(full_data, binary_path)
        except:
            pass
    
    if not from_json:
        marshal_rows = marshal.load(open(binary_path, 'rb'))
        loaded_chemicals = loadChemicalConstants(marshal_rows, rows=True)



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''
import pytest
from thermo.utils import TPDependentProperty
    
def test_tp_dependent_property_exceptions():
    BAD_METHOD = 'BAD_METHOD'
    CRAZY_METHOD = 'CRAZY_METHOD' 
    class MockHeatCapacity(TPDependentProperty):
        name = 'Heat capacity'
        units = 'J/mol'
        ranked_methods_P = [BAD_METHOD, CRAZY_METHOD]
        
        def __init__(self, extrapolation, CASRN, **kwargs):
            self.CASRN = CASRN
            super(MockHeatCapacity, self).__init__(extrapolation, **kwargs)
        
        def load_all_methods(self, load_data):
            if load_data:
                self.T_limits = {BAD_METHOD: (300., 500.),
                                 CRAZY_METHOD: (300, 500)}
                self.all_methods = set()
                self.all_methods_P = {BAD_METHOD, CRAZY_METHOD}
    
        def calculate_P(self, T, P, method):
            if method == BAD_METHOD:
                raise Exception('BAD CALCULATION')
            elif method == CRAZY_METHOD:
                return -1e6
            else:
                return self._base_calculate(T, method)
    
    MVP = MockHeatCapacity(extrapolation='linear', CASRN='7732-18-5')
    MVP.RAISE_PROPERTY_CALCULATION_ERROR = True
    with pytest.raises(RuntimeError):
        try:
            MVP.TP_dependent_property(340., 101325.)
        except RuntimeError as error:
            assert str(error) == ("Failed to evaluate heat capacity method "
                                  "'BAD_METHOD' at T=340.0 K and P=101325.0 Pa "
                                  "for component with CASRN '7732-18-5'")
            raise error
    with pytest.raises(RuntimeError):
        try:
            MVP.TP_dependent_property(520., 101325.)
        except RuntimeError as error:
            assert str(error) == ("Heat capacity method 'BAD_METHOD' is not "
                                  "valid at T=520.0 K and P=101325.0 Pa for "
                                  "component with CASRN '7732-18-5'")
            raise error
    MVP.method_P = CRAZY_METHOD
    with pytest.raises(RuntimeError):
        try:
            MVP.TP_dependent_property(350., 101325.)
        except RuntimeError as error:
            assert str(error) == ("Heat capacity method 'CRAZY_METHOD' "
                                  "computed an invalid value of -1000000.0 J/mol "
                                  "for component with CASRN '7732-18-5'")
            raise error
    MVP.method_P = None
    with pytest.raises(RuntimeError):
        try:
            MVP.TP_dependent_property(340., 101325.)
        except RuntimeError as error:
            assert str(error) == ("No pressure-dependent heat capacity method "
                                  "selected for component with CASRN '7732-18-5'")
            raise error
    


# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2020, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

import pytest
import thermo
from thermo import *
import chemicals
from chemicals import *
from fluids.numerics import *
from math import *
from chemicals.utils import hash_any_primitive
import json
import os
import numpy as np

@pytest.mark.fuzz
@pytest.mark.slow
def test_ChemicalConstantsPackage_from_json_as_json_large():
    create_compounds = []
    for k in dippr_compounds():
        try:
            if search_chemical(k) is not None:
                create_compounds.append(k)
        except:
            pass

    obj = ChemicalConstantsPackage.constants_from_IDs(create_compounds)
    obj2 = ChemicalConstantsPackage.from_json(json.loads(json.dumps(obj.as_json())))

    assert hash(obj) == hash(obj2)
    assert obj == obj2
    assert id(obj) != id(obj2)

    obj = ChemicalConstantsPackage.correlations_from_IDs(create_compounds)
    obj2 = PropertyCorrelationsPackage.from_json(json.loads(json.dumps(obj.as_json())))
    assert hash(obj) == hash(obj2)
    assert obj == obj2
    assert id(obj) != id(obj2)
    
    assert obj != int
    assert obj != float

def test_ChemicalConstantsPackage_json_version_exported():
    constants = ChemicalConstantsPackage(MWs=[18.01528, 106.165], names=['water', 'm-xylene'])
    string = json.dumps(constants.as_json())
    c2 = ChemicalConstantsPackage.from_json(json.loads(string))
    assert 'py/object' in string
    assert 'json_version' in string
    assert not hasattr(c2, 'json_version')

def test_ChemicalConstantsPackage_json_export_does_not_change_hashes():
    # There was a nasty bug where the hashing function was changing its result
    # every call
    obj = ChemicalConstantsPackage.correlations_from_IDs(['hexane'])
    hashes_orig = [hash_any_primitive(getattr(obj, k)) for k in obj.correlations]
    copy = obj.as_json()
    hashes_after = [hash_any_primitive(getattr(obj, k)) for k in obj.correlations]
    assert hashes_orig == hashes_after


def test_ChemicalConstantsPackage_json_export_sane_recursion():

    # It might be nice to do something about the duplicate EOSs, but they could be different
    # Really still feels like a different structure for that would be better.
    obj = ChemicalConstantsPackage.correlations_from_IDs(['methane', 'ethane'])
    assert 3 == json.dumps(obj.as_json()).count('VaporPressure')

def test_ChemicalConstantsPackage_json_export_same_output():
    obj = ChemicalConstantsPackage.correlations_from_IDs(['hexane'])
    obj2 = PropertyCorrelationsPackage.from_json(json.loads(json.dumps(obj.as_json())))

    assert hash_any_primitive(obj.constants) == hash_any_primitive(obj2.constants)
    for prop in obj.pure_correlations:
        assert hash_any_primitive(getattr(obj, prop)) ==  hash_any_primitive(getattr(obj2, prop))
    assert hash_any_primitive(obj.VaporPressures) == hash_any_primitive(obj2.VaporPressures)
    assert hash_any_primitive(obj.ViscosityGases) == hash_any_primitive(obj2.ViscosityGases)
    assert hash(obj.SurfaceTensionMixture) == hash(obj2.SurfaceTensionMixture)
    assert hash(obj.VolumeGasMixture) == hash(obj2.VolumeGasMixture)
    for prop in obj.mixture_correlations:
        assert hash_any_primitive(getattr(obj, prop)) ==  hash_any_primitive(getattr(obj2, prop))


    assert hash(obj) == hash(obj2)
    assert obj == obj2

def test_ChemicalConstantsPackage_wrong_behaviors():
    obj = ChemicalConstantsPackage.correlations_from_IDs(['7647-19-0'])
    obj.VolumeLiquids[0].eos is None
    assert obj != int
    assert obj != float


def test_lemmon2000_package():
    Ts = (150.0, 200.0, 300.0, 1000.0, 2000.0)
    CoolProp_Cps = [29.030484473246823, 29.03511836728048, 29.103801681330573, 33.046833525551676, 36.210748112152906]
    for T, Cp in zip(Ts, CoolProp_Cps):
        assert_close(Cp, lemmon2000_correlations.HeatCapacityGases[0](T), rtol=2e-7)


def test_compound_index():
    obj = ChemicalConstantsPackage(MWs=[18.01528, 106.165], names=['water', 'm-xylene'],
                             CASs=['7732-18-5', '108-38-3'],
                             InChI_Keys=['XLYOFNOQVPJJNP-UHFFFAOYSA-N', 'IVSZLXZYQVIEFR-UHFFFAOYSA-N'], 
                             InChIs=['H2O/h1H2', 'C8H10/c1-7-4-3-5-8(2)6-7/h3-6H,1-2H3'],
                             smiless=['O', 'CC1=CC(=CC=C1)C'], PubChems=[962, 7929],)
    assert 0 == obj.compound_index(name='water')
    assert 1 == obj.compound_index(name='m-xylene')
    assert 1 == obj.compound_index(PubChem=7929)
    assert 0 == obj.compound_index(smiles='O')
    assert 0 == obj.compound_index(CAS='7732-18-5')
    assert 0 == obj.compound_index(InChI='H2O/h1H2')
    assert 1 == obj.compound_index(InChI_Key='IVSZLXZYQVIEFR-UHFFFAOYSA-N')
    
    
def test_add_ChemicalConstantsPackage():
    a = ChemicalConstantsPackage.constants_from_IDs(IDs=['water', 'hexane'])
    b = ChemicalConstantsPackage.constants_from_IDs(IDs=['toluene'])
    c = a + b
    
    c_good = ChemicalConstantsPackage.constants_from_IDs(IDs=['water', 'hexane', 'toluene'])
    assert c == c_good
    
def test_add_PropertyCorrelationsPackage():
    a = ChemicalConstantsPackage.correlations_from_IDs(IDs=['water', 'hexane'])
    b = ChemicalConstantsPackage.correlations_from_IDs(IDs=['toluene'])
    c = a + b
    
    c_good = ChemicalConstantsPackage.correlations_from_IDs(IDs=['water', 'hexane', 'toluene'])
    assert c == c_good


# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2019, 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''
__all__ = ['DryAirLemmon', 'HumidAirRP1485']

from chemicals import air
from chemicals.virial import Z_from_virial_density_form
from chemicals.viscosity import mu_air_lemmon
from chemicals.thermal_conductivity import k_air_lemmon
from .helmholtz_eos import HelmholtzEOS
from .virial_phase import VirialGas
from .iapws_phase import IAPWS95

class DryAirLemmon(HelmholtzEOS):
    model_name = 'lemmon2000'
    is_gas = True
    is_liquid = False
    force_phase = 'g'

    _MW = air.lemmon2000_air_MW
    _MW = 28.96546 # CoolProp
    rho_red = air.lemmon2000_air_rho_reducing
    rho_red_inv = 1.0/rho_red
    T_red = air.lemmon2000_air_T_reducing
    R = air.lemmon2000_air_R

    zs = [1.0]
    cmps = [0]
    N = 1
    T_MAX_FIXED = 2000.0
    T_MIN_FLASH = T_MIN_FIXED = 132.6313 # For now gas only.
    T_fixed_transport = 265.262

    _Ar_func = staticmethod(air.lemmon2000_air_Ar)

    _d3Ar_ddeltadtau2_func = staticmethod(air.lemmon2000_air_d3Ar_ddeltadtau2)
    _d3Ar_ddelta2dtau_func = staticmethod(air.lemmon2000_air_d3Ar_ddelta2dtau)
    _d2Ar_ddeltadtau_func = staticmethod(air.lemmon2000_air_d2Ar_ddeltadtau)

    _dAr_dtau_func = staticmethod(air.lemmon2000_air_dAr_dtau)
    _d2Ar_dtau2_func = staticmethod(air.lemmon2000_air_d2Ar_dtau2)
    _d3Ar_dtau3_func = staticmethod(air.lemmon2000_air_d3Ar_dtau3)
    _d4Ar_dtau4_func = staticmethod(air.lemmon2000_air_d4Ar_dtau4)

    _dAr_ddelta_func = staticmethod(air.lemmon2000_air_dAr_ddelta)
    _d2Ar_ddelta2_func = staticmethod(air.lemmon2000_air_d2Ar_ddelta2)
    _d3Ar_ddelta3_func = staticmethod(air.lemmon2000_air_d3Ar_ddelta3)
    _d4Ar_ddelta4_func = staticmethod(air.lemmon2000_air_d4Ar_ddelta4)

    _d4Ar_ddelta2dtau2_func = staticmethod(air.lemmon2000_air_d4Ar_ddelta2dtau2)
    _d4Ar_ddelta3dtau_func = staticmethod(air.lemmon2000_air_d4Ar_ddelta3dtau)
    _d4Ar_ddeltadtau3_func = staticmethod(air.lemmon2000_air_d4Ar_ddeltadtau3)

    def __init__(self, T=None, P=None, zs=None):
        self.T = T
        self.P = P
        self._rho = rho = air.lemmon2000_rho(T, P)
        self._V = 1.0/rho
        self.tau = tau = self.T_red/T
        self.delta = delta = rho*self.rho_red_inv
        self.A0 = air.lemmon2000_air_A0(tau, delta)
        self.dA0_dtau = air.lemmon2000_air_dA0_dtau(tau, delta)
        self.d2A0_dtau2 = air.lemmon2000_air_d2A0_dtau2(tau, delta)
        self.d3A0_dtau3 = air.lemmon2000_air_d3A0_dtau3(tau, delta)

    def to_TP_zs(self, T, P, zs):
        new = self.__class__.__new__(self.__class__)
        new.zs = zs
        new.T = T
        new.P = P
        new._rho = rho = air.lemmon2000_rho(T, P)
        new._V = 1.0/rho
        new.tau = tau = new.T_red/T
        new.delta = delta = rho*new.rho_red_inv
        new.A0 = air.lemmon2000_air_A0(tau, delta)
        new.dA0_dtau = air.lemmon2000_air_dA0_dtau(tau, delta)
        new.d2A0_dtau2 = air.lemmon2000_air_d2A0_dtau2(tau, delta)
        new.d3A0_dtau3 = air.lemmon2000_air_d3A0_dtau3(tau, delta)
        return new

    def to(self, zs, T=None, P=None, V=None):
        new = self.__class__.__new__(self.__class__)
        new.zs = zs
        if T is not None and P is not None:
            new.T = T
            new._rho = air.lemmon2000_rho(T, P)
            new._V = 1.0/new._rho
            new.P = P
        elif T is not None and V is not None:
            new._rho = 1.0/V
            new._V = V
            P = air.lemmon2000_P(T, new._rho)
        elif P is not None and V is not None:
            T = air.lemmon2000_T(P=P, rho=1.0/V)
            new._rho = 1.0/V
            new._V = V
        else:
            raise ValueError("Two of T, P, or V are needed")

        new.P = P
        new.T = T
        new.tau = tau = new.T_red/T
        new.delta = delta = new._rho*new.rho_red_inv

        new.A0 = air.lemmon2000_air_A0(tau, delta)
        new.dA0_dtau = air.lemmon2000_air_dA0_dtau(tau, delta)
        new.d2A0_dtau2 = air.lemmon2000_air_d2A0_dtau2(tau, delta)
        new.d3A0_dtau3 = air.lemmon2000_air_d3A0_dtau3(tau, delta)
        return new

    def mu(self):
        r'''Calculate and return the viscosity of air according to the Lemmon
        and Jacobsen (2003) .
        For details, see :obj:`chemicals.viscosity.mu_air_lemmon`.

        Returns
        -------
        mu : float
            Viscosity of air, [Pa*s]
        '''
        try:
            return self._mu
        except:
            pass
        self._mu = mu = mu_air_lemmon(self.T, self._rho)
        return mu

    def k(self):
        r'''Calculate and return the thermal conductivity of air according to 
        Lemmon and Jacobsen (2004)
        For details, see :obj:`chemicals.thermal_conductivity.k_air_lemmon`.

        Returns
        -------
        k : float
            Thermal conductivity of air, [W/m/K]
        '''
        try:
            return self._k
        except:
            pass
        # We require viscosity to calculate thermal conductivity
        self.mu()
        
        # This call is very expensive; this could be curve-fit as it is a 1D function
        drho_dP_Tr = self.to(T=self.T_fixed_transport, V=self._V, zs=self.zs).drho_dP()

        self._k = k = k_air_lemmon(T=self.T, rho=self._rho, Cp=self.Cp(), Cv=self.Cv(),
                       mu=self._mu, drho_dP=self.drho_dP(), drho_dP_Tr=drho_dP_Tr)
        return k



class HumidAirRP1485(VirialGas):
    is_gas = True
    is_liquid = False
    def __init__(self, Hfs=None, Gfs=None, T=None, P=None, zs=None,
                 ):
        # Although in put is zs, it is required to be in the order of
        # (air, water) mole fraction
        self.Hfs = Hfs
        self.Gfs = Gfs
        if Hfs is not None and Gfs is not None and None not in Hfs and None not in Gfs:
            self.Sfs = [(Hfi - Gfi)/298.15 for Hfi, Gfi in zip(Hfs, Gfs)]
        else:
            self.Sfs = None

        for i in (Hfs, Gfs, zs):
            if i is not None:
                self.N = len(i)
                break
        if zs is not None:
            self.psi_w = psi_w = zs[1]
            self.psi_a = psi_a = zs[0]
            self.zs = zs
        if T is not None:
            self.T = T
        if P is not None:
            self.P = P
        if T is not None and P is not None and zs is not None:
            self.air = DryAirLemmon(T=T, P=P)
            self.water = IAPWS95(T=T, P=P)
            Z = Z_from_virial_density_form(T, P, self.B(), self.C())
            self._V = Z*self.R*T/P
            self._MW = DryAirLemmon._MW*psi_a + IAPWS95._MW*psi_w

    def B(self):
        try:
            return self._B
        except:
            pass
        Baa = self.air.B_virial()
        Baw = air.TEOS10_BAW_derivatives(self.T)[0]
        Bww = self.water.B_virial()
        psi_a, psi_w = self.psi_a, self.psi_w

        self._B = B = psi_a*psi_a*Baa + 2.0*psi_a*psi_w*Baw + psi_w*psi_w*Bww
        return B

    def C(self):
        try:
            return self._C
        except:
            pass
        T = self.T
        Caaa = self.air.C_virial()
        Cwww = self.water.C_virial()
        Caww = air.TEOS10_CAWW_derivatives(T)[0]
        Caaw = air.TEOS10_CAAW_derivatives(T)[0]
        psi_a, psi_w = self.psi_a, self.psi_w
        self._C = C = (psi_a*psi_a*(Caaa + 3.0*psi_w*Caaw)
                       + psi_w*psi_w*(3.0*psi_a*Caww + psi_w*Cwww))
        return C



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2019, 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''
__all__ = ['HelmholtzEOS',]

from chemicals.utils import log
from .phase import Phase

class HelmholtzEOS(Phase):

    model_attributes = ('model_name',)

    def __repr__(self):
        r'''Method to create a string representation of the phase object, with
        the goal of making it easy to obtain standalone code which reproduces
        the current state of the phase. This is extremely helpful in creating
        new test cases.

        Returns
        -------
        recreation : str
            String which is valid Python and recreates the current state of
            the object if ran, [-]

        Examples
        --------
        >>> from thermo import IAPWS95Gas
        >>> phase = IAPWS95Gas(T=300, P=1e5, zs=[1])
        >>> phase
        IAPWS95Gas(T=300, P=100000.0, zs=[1.0])
        '''
        base = '%s('  %(self.__class__.__name__)
        for s in ('Hfs', 'Gfs', 'Sfs', 'T', 'P', 'zs'):
            if hasattr(self, s) and getattr(self, s) is not None:
                base += '%s=%s, ' %(s, getattr(self, s))
        if base[-2:] == ', ':
            base = base[:-2]
        base += ')'
        return base

    def V(self):
        return self._V

    def A(self):
        try:
            return self._A
        except:
            pass
        A = self._A = self.A0 + self._Ar_func(self.tau, self.delta)
        return A

    def dA_ddelta(self):
        try:
            return self._dA_ddelta
        except:
            pass
        delta = self.delta
        dA_ddelta = self._dA_ddelta = self._dAr_ddelta_func(self.tau, delta) + 1./delta
        return dA_ddelta

    def d2A_ddelta2(self):
        try:
            return self._d2A_ddelta2
        except:
            pass
        delta = self.delta
        self._d2A_ddelta2 = d2A_ddelta2 = self._d2Ar_ddelta2_func(self.tau, delta) - 1./(delta*delta)
        return d2A_ddelta2


    def d3A_ddelta3(self):
        try:
            return self._d3A_ddelta3
        except:
            pass
        delta = self.delta
        self._d3A_ddelta3 = d3A_ddelta3 = self._d3Ar_ddelta3_func(self.tau, delta) + 2./(delta*delta*delta)
        return d3A_ddelta3

    def dA_dtau(self):
        try:
            return self._dA_dtau
        except:
            pass
        dA_dtau = self._dA_dtau = self._dAr_dtau_func(self.tau, self.delta) + self.dA0_dtau
        return dA_dtau

    def d2A_dtau2(self):
        try:
            return self._d2A_dtau2
        except:
            pass
        self._d2A_dtau2 = d2A_dtau2 = self._d2Ar_dtau2_func(self.tau, self.delta) + self.d2A0_dtau2
        return d2A_dtau2

    def d2A_ddeltadtau(self):
        try:
            return self._d2A_ddeltadtau
        except:
            pass
        self._d2A_ddeltadtau = d2A_ddeltadtau = self._d2Ar_ddeltadtau_func(self.tau, self.delta)
        return d2A_ddeltadtau

    def d3A_ddelta2dtau(self):
        try:
            return self._d3A_ddelta2dtau
        except:
            pass
        self._d3A_ddelta2dtau = d3A_ddelta2dtau = self._d3Ar_ddelta2dtau_func(self.tau, self.delta)
        return d3A_ddelta2dtau

    def d3A_ddeltadtau2(self):
        try:
            return self._d3A_ddeltadtau2
        except:
            pass
        self._d3A_ddeltadtau2 = d3A_ddeltadtau2 = self._d3Ar_ddeltadtau2_func(self.tau, self.delta)
        return d3A_ddeltadtau2

    def lnphis(self):
        delta = self.delta
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            A = self._A
        except:
            A = self.A()
        x0 = delta*(dA_ddelta - 1.0/delta)
        lnphi = A - self.A0 + x0 - log(x0 + 1.0)

        return [lnphi]

    def dlnphis_dV_T(self):
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            d2A_ddelta2 = self._d2A_ddelta2
        except:
            d2A_ddelta2 = self.d2A_ddelta2()
        delta = self.delta
        rho, rho_red = 1.0/self._V, self.rho_red
        x0 = self.rho_red_inv
        rho_inv = 1.0/rho

        x1 = -rho_red*rho_inv
        x2 = rho*(d2A_ddelta2*x0 + rho_red*rho_inv*rho_inv)
        x3 = dA_ddelta + x1
        dlnphi_dV_T = x0*(2.0*dA_ddelta + x1 + x2 - (x2 + x3)/(rho*x0*x3 + 1.0) - 1.0/delta)
#
        dlnphi_dV_T *= -1.0/(self._V*self._V)

        return [dlnphi_dV_T]

    def dlnphis_dT_V(self):
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            d2A_ddelta2 = self._d2A_ddelta2
        except:
            d2A_ddelta2 = self.d2A_ddelta2()
        try:
            d2A_ddeltadtau = self._d2A_ddeltadtau
        except:
            d2A_ddeltadtau = self.d2A_ddeltadtau()
        try:
            dA_dtau = self._dA_dtau
        except:
            dA_dtau = self.dA_dtau()

        rho = 1.0/self._V
        rho_red_inv = self.rho_red_inv
        rho_red = self.rho_red

        T_red, T = self.T_red, self.T

        dlnphi_dT_V = T_red/(T*T)*(rho_red_inv*rho*d2A_ddeltadtau*(-1.0
                        + 1.0/(rho*(dA_ddelta - rho_red/rho)*rho_red_inv + 1.0))
                        - dA_dtau
                        + self.dA0_dtau)
        return [dlnphi_dT_V]

    def dlnphis_dT_P(self):
        try:
            return self._dlnphis_dT_P
        except:
            pass

        self._dlnphis_dT_P = dlnphis_dT_P = [self.dlnphis_dT_V()[0] - self.dlnphis_dV_T()[0]*self.dP_dT_V()/self.dP_dV_T()]
        return dlnphis_dT_P
    dlnphis_dT = dlnphis_dT_P

    def dlnphis_dP_V(self):
        return [self.dlnphis_dT_V()[0]/self.dP_dT_V()]

    def dlnphis_dV_P(self):
        return [self.dlnphis_dT_P()[0]/self.dV_dT_P()]

    def dlnphis_dP_T(self):
        return [self.dlnphis_dP_V()[0] - self.dlnphis_dV_P()[0]*self.dT_dP_V()/self.dT_dV_P()]

    dlnphis_dP = dlnphis_dP_T

    def S(self):
        try:
            return self._S
        except:
            pass
        try:
            dA_dtau = self._dA_dtau
        except:
            dA_dtau = self.dA_dtau()
        try:
            A = self._A
        except:
            A = self.A()
        self._S = S = (self.tau*dA_dtau - A)*self.R
        return S

    def S_dep(self):
        S_ideal = (self.tau*self.dA0_dtau - self.A0)*self.R
        return self.S() - S_ideal


    def dS_dT_V(self):
        try:
            return self._dS_dT_V
        except:
            pass
        try:
            d2A_dtau2 = self._d2A_dtau2
        except:
            d2A_dtau2 = self.d2A_dtau2()

        T, T_red = self.T, self.T_red
        self._dS_dT_V = dS_dT_V = (-T_red*T_red*d2A_dtau2)*self.R/(T*T*T)
        return dS_dT_V

    def dS_dV_T(self):
        try:
            return self._dS_dV_T
        except:
            pass
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            d2A_ddeltadtau = self._d2A_ddeltadtau
        except:
            d2A_ddeltadtau = self.d2A_ddeltadtau()

        R, T, V, rho_red, T_red = self.R, self.T, self._V, self.rho_red, self.T_red

        self._dS_dV_T = dS_dV_T = R*(dA_ddelta - T_red*d2A_ddeltadtau/(T))/(V*V*rho_red)
        return dS_dV_T

    def dS_dP_T(self):
        # From chain rule
        return self.dS_dV_T()/self.dP_dV()
    dS_dP = dS_dP_T


    def dS_dT_P(self):
        try:
            return self._dS_dT_P
        except:
            pass

        self._dS_dT_P = dS_dT_P = self.dS_dT_V() - self.dS_dV_T()*self.dP_dT_V()/self.dP_dV_T()
        return dS_dT_P
    dS_dT = dS_dT_P

    def dS_dP_V(self):
        return self.dS_dT_V()/self.dP_dT_V()

    def H(self):
        try:
            return self._H
        except:
            pass
        try:
            dA_dtau = self._dA_dtau
        except:
            dA_dtau = self.dA_dtau()
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        self._H = H = (self.tau*dA_dtau + dA_ddelta*self.delta)*self.T*self.R
        return H

    def H_dep(self):
        try:
            return self._H_dep
        except:
            pass
        # Might be right, might not - hard to check due to reference state
        H = self.H()
        dA_dtau = self.dA0_dtau
        dA_ddelta = 1./self.delta

        H_ideal = (self.tau*dA_dtau + dA_ddelta*self.delta)*self.T*self.R
        self._H_dep = H - H_ideal
        return self._H_dep

    def dH_dT_V(self):
        try:
            return self._dH_dT_V
        except:
            pass
        try:
            dA_dtau = self._dA_dtau
        except:
            dA_dtau = self.dA_dtau()
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            d2A_ddeltadtau = self._d2A_ddeltadtau
        except:
            d2A_ddeltadtau = self.d2A_ddeltadtau()
        try:
            d2A_dtau2 = self._d2A_dtau2
        except:
            d2A_dtau2 = self.d2A_dtau2()


        T, T_red, V, R, rho_red = self.T, self.T_red, self._V, self.R, self.rho_red
        T_inv = 1.0/T

        self._dH_dT_V = dH_dT_V = R*T*(-T_red*dA_dtau*T_inv*T_inv - T_inv*T_inv*T_red*d2A_ddeltadtau/(V*rho_red) - T_red*T_red*T_inv*T_inv*T_inv*d2A_dtau2) + R*(dA_ddelta/(V*rho_red) + T_red*dA_dtau/T)
        return dH_dT_V

    def dH_dV_T(self):
        try:
            return self._dH_dV_T
        except:
            pass
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            d2A_ddeltadtau = self._d2A_ddeltadtau
        except:
            d2A_ddeltadtau = self.d2A_ddeltadtau()
        try:
            d2A_ddelta2 = self._d2A_ddelta2
        except:
            d2A_ddelta2 = self.d2A_ddelta2()

        T, T_red, V, R, rho_red = self.T, self.T_red, self._V, self.R, self.rho_red

        self._dH_dV_T = dH_dV_T = R*T*(-dA_ddelta/(V*V*rho_red) - d2A_ddelta2/(V*V*V*rho_red*rho_red) - T_red*d2A_ddeltadtau/(T*V*V*rho_red))
        return dH_dV_T

    def dH_dP_V(self):
        return self.dH_dT_V()/self.dP_dT_V()

    def dH_dP_T(self):
        # From chain rule
        return self.dH_dV_T()/self.dP_dV()
    dH_dP = dH_dP_T

    def dH_dV_P(self):
        return self.dH_dT_P()/self.dV_dT_P()



    def Cv(self):
        try:
            return self._Cv
        except:
            pass
        try:
            d2A_dtau2 = self._d2A_dtau2
        except:
            d2A_dtau2 = self.d2A_dtau2()
        tau = self.tau
        self._Cv = Cv = -tau*tau*d2A_dtau2*self.R
        return Cv

    def Cp(self):
        try:
            return self._Cp
        except:
            pass
        tau, delta = self.tau, self.delta
        try:
            d2A_dtau2 = self._d2A_dtau2
        except:
            d2A_dtau2 = self.d2A_dtau2()
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            d2A_ddelta2 = self._d2A_ddelta2
        except:
            d2A_ddelta2 = self.d2A_ddelta2()
        try:
            d2A_ddeltadtau = self._d2A_ddeltadtau
        except:
            d2A_ddeltadtau = self.d2A_ddeltadtau()

        x0 = delta*dA_ddelta - delta*tau*d2A_ddeltadtau
        den = delta*(dA_ddelta + dA_ddelta + delta*d2A_ddelta2)
        self._Cp = Cp = (-tau*tau*d2A_dtau2 + x0*x0/den)*self.R
        return Cp

    dH_dT = dH_dT_P = Cp


    def dP_dT(self):
        try:
            return self._dP_dT
        except:
            pass
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            d2A_ddeltadtau = self._d2A_ddeltadtau
        except:
            d2A_ddeltadtau = self.d2A_ddeltadtau()
        self._dP_dT = dP_dT = self.R/self._V*self.delta*(dA_ddelta - self.tau*d2A_ddeltadtau)
        return dP_dT

    def dP_dV(self):
        try:
            return self._dP_dV
        except:
            pass
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            d2A_ddelta2 = self._d2A_ddelta2
        except:
            d2A_ddelta2 = self.d2A_ddelta2()
        delta = self.delta

        rho = self.rho()
        self._dP_dV = dP_dV = -self.R*rho*rho*self.T*delta*(2.0*dA_ddelta + delta*d2A_ddelta2)
        return dP_dV

    dP_dT_V = dP_dT
    dP_dV_T = dP_dV

    def d2P_dT2(self):
        try:
            return self._d2P_dT2
        except:
            pass
        d3A_ddeltadtau2 = self.d3A_ddeltadtau2()
        T, T_red, delta, V, rho_red, R = self.T, self.T_red, self.delta, self._V, self.rho_red, self.R
        self._d2P_dT2 = d2P_dT2 = R*T_red*T_red*d3A_ddeltadtau2/(T*T*T*V*V*rho_red)
        return d2P_dT2

    def d2P_dV2(self):
        try:
            return self._d2P_dV2
        except:
            pass
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            d2A_ddelta2 = self._d2A_ddelta2
        except:
            d2A_ddelta2 = self.d2A_ddelta2()
        try:
            d3A_ddelta3 = self._d3A_ddelta3
        except:
            d3A_ddelta3 = self.d3A_ddelta3()
        '''
        from sympy import *
        from chemicals import Vm_to_rho
        T_red, rho_red, V, R = symbols('T_red, rho_red, V, R')
        T = symbols('T')
        rho = 1/V

        iapws95_dA_ddelta = symbols('ddelta', cls=Function)

        rho_red_inv = (1/rho_red)
        tau = T_red/T
        delta = rho*rho_red_inv
        dA_ddelta = iapws95_dA_ddelta(tau, delta)
        P = (dA_ddelta*delta)*rho*R*T
        print(diff(P,V, 2))
        '''
        R, T, V, rho_red = self.R, self.T, self._V, self.rho_red

        self._d2P_dV2 = d2P_dV2 = R*T*(6.0*dA_ddelta + (2*d2A_ddelta2 + d3A_ddelta3/(V*rho_red))/(V*rho_red) + 4*d2A_ddelta2/(V*rho_red))/(V**4*rho_red)
        return d2P_dV2

    def d2P_dTdV(self):
        try:
            return self._d2P_dTdV
        except:
            pass
        try:
            dA_ddelta = self._dA_ddelta
        except:
            dA_ddelta = self.dA_ddelta()
        try:
            d2A_ddelta2 = self._d2A_ddelta2
        except:
            d2A_ddelta2 = self.d2A_ddelta2()
        try:
            d2A_ddeltadtau = self._d2A_ddeltadtau
        except:
            d2A_ddeltadtau = self.d2A_ddeltadtau()
        d3A_ddelta2dtau = self.d3A_ddelta2dtau()
        R, T, T_red, V, rho_red = self.R, self.T, self.T_red, self._V, self.rho_red

        self._d2P_dTdV = d2P_dTdV = R*(-2.0*dA_ddelta - d2A_ddelta2/(V*rho_red) + 2.0*T_red*d2A_ddeltadtau/T
                                       + T_red*d3A_ddelta2dtau/(T*V*rho_red))/(V*V*V*rho_red)
        return d2P_dTdV

    def B_virial(self):
        try:
            f0 = self._dAr_ddelta_func(self.tau, 0.0)
        except:
            f0 = self._dAr_ddelta_func(self.tau, 1e-20)
        return f0/self.rho_red

    def dB_virial_dT(self):
        tau = self.tau
        try:
            f0 = self._d2Ar_ddeltadtau_func(tau, 0.0)
        except:
            f0 = self._d2Ar_ddeltadtau_func(tau, 1e-20)
        return -tau*tau*f0/(self.rho_red*self.T_red)

    def d2B_virial_dT2(self):
        T, T_red, tau = self.T, self.T_red, self.tau

        try:
            f0 = self._d2Ar_ddeltadtau_func(tau, 0.0)
        except:
            f0 = self._d2Ar_ddeltadtau_func(tau, 1e-20)
        try:
            f1 = self._d3Ar_ddeltadtau2_func(tau, 0.0)
        except:
            f1 = self._d3Ar_ddeltadtau2_func(tau, 1e-20)
        return T_red*(2.0*f0 + tau*f1)/(T*T*T*self.rho_red)

    def d3B_virial_dT3(self):
        T, T_red, tau = self.T, self.T_red, self.tau

        try:
            f0 = self._d2Ar_ddeltadtau_func(tau, 0.0)
        except:
            f0 = self._d2Ar_ddeltadtau_func(tau, 1e-20)
        try:
            f1 = self._d3Ar_ddeltadtau2_func(tau, 0.0)
        except:
            f1 = self._d3Ar_ddeltadtau2_func(tau, 1e-20)
        try:
            f2 = self._d4Ar_ddeltadtau3_func(tau, 0.0)
        except:
            f2 = self._d4Ar_ddeltadtau3_func(tau, 1e-20)
        return (-T_red*(6.0*f0 + 6.0*tau*f1 + tau*tau*f2)/(T*T*T*T*self.rho_red))

    def C_virial(self):
        try:
            f0 = self._d2Ar_ddelta2_func(self.tau, 0.0)
        except:
            f0 = self._d2Ar_ddelta2_func(self.tau, 1e-20)
        return f0/(self.rho_red*self.rho_red)

    def dC_virial_dT(self):
        tau = self.tau
        try:
            f0 = self._d3Ar_ddelta2dtau_func(tau, 0.0)
        except:
            f0 = self._d3Ar_ddelta2dtau_func(tau, 1e-20)
        return -tau*tau*f0/(self.rho_red*self.rho_red*self.T_red)

    def d2C_virial_dT2(self):
        T, T_red, tau = self.T, self.T_red, self.tau
        try:
            f0 = self._d3Ar_ddelta2dtau_func(tau, 0.0)
        except:
            f0 = self._d3Ar_ddelta2dtau_func(tau, 1e-20)

        try:
            f1 = self._d4Ar_ddelta2dtau2_func(tau, 0.0)
        except:
            f1 = self._d4Ar_ddelta2dtau2_func(tau, 1e-20)
        return T_red*(2.0*f0 + tau*f1)/(T*T*T*self.rho_red*self.rho_red)




# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2019, 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This module contains functions for identifying phases as liquid, solid, and
gas.

Solid identification is easy using the
:obj:`phase identification parameter <chemicals.utils.phase_identification_parameter_phase>`.
There is never more than one gas by definition. For pure species, the
phase identification parameter is a clear vapor-liquid differentiator in the
subcritical region and it provides line starting at the critical point for the
supercritical region.

However for mixtures, there is no clear calcuation that can be performed to
identify the phase of a mixture. Many different criteria that have been
proposed are included here. The
:obj:`phase identification parameter or PIP <chemicals.utils.phase_identification_parameter_phase>`.
is recommended in general and is the default.

For reporting bugs, adding feature requests, or submitting pull requests,
please use the `GitHub issue tracker <https://github.com/CalebBell/thermo/>`_.

.. contents:: :local:

Phase Identification
====================

Main Interface
--------------
.. autofunction:: identify_sort_phases

Secondary Interfaces
--------------------
.. autofunction:: identity_phase_states
.. autodata:: VL_ID_METHODS
.. autodata:: S_ID_METHODS

Scoring Functions
-----------------
.. autofunction:: score_phases_VL
.. autofunction:: score_phases_S
.. autofunction:: vapor_score_traces
.. autofunction:: vapor_score_Tpc
.. autofunction:: vapor_score_Vpc
.. autofunction:: vapor_score_Tpc_weighted
.. autofunction:: vapor_score_Tpc_Vpc
.. autofunction:: vapor_score_Wilson
.. autofunction:: vapor_score_Poling
.. autofunction:: vapor_score_PIP
.. autofunction:: vapor_score_Bennett_Schmidt

Sorting Phases
==============
.. autofunction:: sort_phases

'''

from __future__ import division
__all__ = ['vapor_score_Tpc', 'vapor_score_Vpc',
           'vapor_score_Tpc_weighted', 'vapor_score_Tpc_Vpc',
           'vapor_score_Wilson', 'vapor_score_Poling',
           'vapor_score_PIP', 'vapor_score_Bennett_Schmidt',
           'vapor_score_traces',

           'score_phases_S', 'score_phases_VL', 'identity_phase_states',
           'S_ID_METHODS', 'VL_ID_METHODS',

           'sort_phases', 'identify_sort_phases',

           'WATER_FIRST', 'WATER_LAST', 'WATER_NOT_SPECIAL',
           'WATER_SORT_METHODS', 'KEY_COMPONENTS_SORT', 'PROP_SORT',
           'SOLID_SORT_METHODS', 'LIQUID_SORT_METHODS',

           'VL_ID_TPC', 'VL_ID_VPC', 'VL_ID_TPC_VC_WEIGHTED', 'VL_ID_TPC_VPC',
           'VL_ID_WILSON', 'VL_ID_POLING', 'VL_ID_PIP', 'VL_ID_BS', 'VL_ID_TRACES',
           'VL_ID_METHODS', 'S_ID_D2P_DVDT', 'S_ID_METHODS',
           ]


from chemicals.rachford_rice import Rachford_Rice_flash_error, flash_inner_loop
from chemicals.flash_basic import Wilson_K_value
from chemicals.utils import phase_identification_parameter, Vm_to_rho

def vapor_score_Tpc(T, Tcs, zs):
    r'''Compute a vapor score representing how vapor-like a phase is
    (higher, above zero = more vapor like) using the following criteria

    .. math::
        T - \sum_i z_i T_{c,i}

    Parameters
    ----------
    T : float
        Temperature, [K]
    Tcs : list[float]
        Critical temperatures of all species, [K]
    zs : list[float]
        Mole fractions of the phase being identified, [-]

    Returns
    -------
    score : float
        Vapor like score, [-]

    Examples
    --------
    A flash of equimolar CO2/n-hexane at 300 K and 1 MPa is computed, and there
    is a two phase solution. The phase must be identified for each result:

    Liquid-like phase:

    >>> vapor_score_Tpc(T=300.0, Tcs=[304.2, 507.6], zs=[0.21834418746784942, 0.7816558125321506])
    -163.18879226903942

    Vapor-like phase:

    >>> vapor_score_Tpc(T=300.0, Tcs=[304.2, 507.6], zs=[0.9752234962374878, 0.024776503762512052])
    -9.239540865294941

    In this result, the vapor phase is not identified as a gas at all! It has
    a mass density of ~ 20 kg/m^3, which would usually be called a gas by most
    people.

    Notes
    -----
    '''
    # Does not work for pure compounds
    Tpc =  0.0
    for i in range(len(zs)):
        Tpc += zs[i]*Tcs[i]
    return T - Tpc

def vapor_score_Vpc(V, Vcs, zs):
    r'''Compute a vapor score representing how vapor-like a phase is
    (higher, above zero = more vapor like) using the following criteria

    .. math::
        V - \sum_i z_i V_{c,i}

    Parameters
    ----------
    V : float
        Molar volume, [m^3/mol]
    Vcs : list[float]
        Critical molar volumes of all species, [m^3/mol]
    zs : list[float]
        Mole fractions of the phase being identified, [-]

    Returns
    -------
    score : float
        Vapor like score, [-]

    Examples
    --------
    A flash of equimolar CO2/n-hexane at 300 K and 1 MPa is computed, and there
    is a two phase solution. The phase must be identified for each result:

    Liquid-like phase:

    >>> vapor_score_Vpc(V=0.00011316308855449715, Vcs=[9.4e-05, 0.000368], zs=[0.21834418746784942, 0.7816558125321506])
    -0.000195010604079

    Vapor-like phase:

    >>> vapor_score_Vpc(V=0.0023406573328250335, Vcs=[9.4e-05, 0.000368], zs=[0.9752234962374878, 0.024776503762512052])
    0.002239868570

    Notes
    -----
    '''
    Vpc =  0.0
    for i in range(len(zs)):
        Vpc += zs[i]*Vcs[i]
    return V - Vpc

def vapor_score_Tpc_weighted(T, Tcs, Vcs, zs, r1=1.0):
    r'''Compute a vapor score representing how vapor-like a phase is
    (higher, above zero = more vapor like) using the following criteria, said
    to be implemented in ECLIPSE [1]_:

    .. math::
        T - T_{pc}

    .. math::
        T_{p,c} = r_1 \frac{\sum_j x_j V_{c,j}T_{c,j}}{\sum_j x_j V_{c,j}}

    Parameters
    ----------
    T : float
        Temperature, [K]
    Tcs : list[float]
        Critical temperatures of all species, [K]
    Vcs : list[float]
        Critical molar volumes of all species, [m^3/mol]
    zs : list[float]
        Mole fractions of the phase being identified, [-]
    r1 : float
        Tuning factor, [-]

    Returns
    -------
    score : float
        Vapor like score, [-]

    Examples
    --------
    A flash of equimolar CO2/n-hexane at 300 K and 1 MPa is computed, and there
    is a two phase solution. The phase must be identified for each result:

    Liquid-like phase:

    >>> vapor_score_Tpc_weighted(T=300.0, Tcs=[304.2, 507.6], Vcs=[9.4e-05, 0.000368], zs=[0.21834418746784942, 0.7816558125321506])
    -194.0535694431

    Vapor-like phase:

    >>> vapor_score_Tpc_weighted(T=300.0, Tcs=[304.2, 507.6], Vcs=[9.4e-05, 0.000368], zs=[0.9752234962374878, 0.024776503762512052])
    -22.60037521107

    As can be seen, the CO2-phase is incorrectly identified as a liquid.

    Notes
    -----

    References
    ----------
    .. [1] Bennett, Jim, and Kurt A. G. Schmidt. "Comparison of Phase
       Identification Methods Used in Oil Industry Flow Simulations." Energy &
       Fuels 31, no. 4 (April 20, 2017): 3370-79.
       https://doi.org/10.1021/acs.energyfuels.6b02316.
    '''
    weight_sum = 0.0
    for i in range(len(zs)):
        weight_sum += zs[i]*Vcs[i]

    Tpc = 0.0
    for i in range(len(zs)):
        Tpc += zs[i]*Tcs[i]*Vcs[i]
    Tpc *= r1/weight_sum

    return T - Tpc

def vapor_score_Tpc_Vpc(T, V, Tcs, Vcs, zs):
    r'''Compute a vapor score representing how vapor-like a phase is
    (higher, above zero = more vapor like) using the following criteria, said
    to be implemented in Multiflash [1]_:

    .. math::
        VT^2 - V_{pc} T_{pc}^2

    Parameters
    ----------
    T : float
        Temperature, [K]
    V : float
        Molar volume, [m^3/mol]
    Tcs : list[float]
        Critical temperatures of all species, [K]
    Vcs : list[float]
        Critical molar volumes of all species, [m^3/mol]
    zs : list[float]
        Mole fractions of the phase being identified, [-]

    Returns
    -------
    score : float
        Vapor like score, [-]

    Examples
    --------
    A flash of equimolar CO2/n-hexane at 300 K and 1 MPa is computed, and there
    is a two phase solution. The phase must be identified for each result:

    Liquid-like phase:

    >>> vapor_score_Tpc_Vpc(T=300.0, V=0.00011316308855449715, Tcs=[304.2, 507.6], Vcs=[9.4e-05, 0.000368], zs=[0.21834418746784942, 0.7816558125321506])
    -55.932094761

    Vapor-like phase:

    >>> vapor_score_Tpc_Vpc(T=300.0, V=0.0023406573328250335, Tcs=[304.2, 507.6], Vcs=[9.4e-05, 0.000368], zs=[0.9752234962374878, 0.024776503762512052])
    201.020821992

    Notes
    -----

    References
    ----------
    .. [1] Bennett, Jim, and Kurt A. G. Schmidt. "Comparison of Phase
       Identification Methods Used in Oil Industry Flow Simulations." Energy &
       Fuels 31, no. 4 (April 20, 2017): 3370-79.
       https://doi.org/10.1021/acs.energyfuels.6b02316.
    '''
    # Basic. Different mixing rules could be used to tune the system.
    Tpc =  0.0
    for i in range(len(zs)):
        Tpc += zs[i]*Tcs[i]
    Vpc =  0.0
    for i in range(len(zs)):
        Vpc += zs[i]*Vcs[i]
    return V*T*T - Vpc*Tpc*Tpc


def vapor_score_Wilson(T, P, zs, Tcs, Pcs, omegas):
    r'''Compute a vapor score representing how vapor-like a phase is
    (higher, above zero = more vapor like) using the Rachford-Rice Wilson
    method of Perschke [1]_.

    After calculating Wilson's K values, the following expression is evaluated
    at :math:`\frac{V}{F} = 0.5`; the result is the score.

    .. math::
        \sum_i \frac{z_i(K_i-1)}{1 + \frac{V}{F}(K_i-1)}

    Parameters
    ----------
    T : float
        Temperature, [K]
    P : float
        Pressure, [Pa]
    zs : list[float]
        Mole fractions of the phase being identified, [-]
    Tcs : list[float]
        Critical temperatures of all species, [K]
    Pcs : list[float]
        Critical pressures of all species, [Pa]
    omegas : list[float]
        Acentric factors of all species, [-]

    Returns
    -------
    score : float
        Vapor like score, [-]

    Examples
    --------
    A flash of equimolar CO2/n-hexane at 300 K and 1 MPa is computed, and there
    is a two phase solution. The phase must be identified for each result:

    Liquid-like phase:

    >>> vapor_score_Wilson(T=300.0, P=1e6, zs=[.218, .782], Tcs=[304.2, 507.6], Pcs=[7376460.0, 3025000.0], omegas=[0.2252, 0.2975])
    -1.16644793

    Vapor-like phase:

    >>> vapor_score_Wilson(T=300.0, P=1e6, zs=[.975, .025], Tcs=[304.2, 507.6], Pcs=[7376460.0, 3025000.0], omegas=[0.2252, 0.2975])
    1.397678492

    This method works well in many conditions, like the Wilson equation itself,
    but fundamentally it cannot do a great job because it is not tied to the
    phase model itself.

    A dew point flash at P = 100 Pa for the same mixture shows both phases
    being identified as vapor-like:

    >>> T_dew = 206.40935716944634
    >>> P = 100.0
    >>> vapor_score_Wilson(T=T_dew, P=P, zs=[0.5, 0.5], Tcs=[304.2, 507.6], Pcs=[7376460.0, 3025000.0], omegas=[0.2252, 0.2975])
    1.074361930956633
    >>> vapor_score_Wilson(T=T_dew, P=P, zs=[0.00014597910182360052, 0.9998540208981763], Tcs=[304.2, 507.6], Pcs=[7376460.0, 3025000.0], omegas=[0.2252, 0.2975])
    0.15021784286075726

    Notes
    -----

    References
    ----------
    .. [1] Chang, Yih-Bor. "Development and Application of an Equation of State
       Compositional Simulator," 1990.
       https://repositories.lib.utexas.edu/handle/2152/80585.
    '''
    N = len(zs)
    if N == 1:
        Psat = Wilson_K_value(T, P, Tcs[0], Pcs[0], omegas[0])*P
        # Lower than vapor pressure - gas; higher than the vapor pressure - liquid
        return P - Psat
    # Does not work for pure compounds
    # Posivie - vapor, negative - liquid
    Ks = [Wilson_K_value(T, P, Tcs[i], Pcs[i], omegas[i]) for i in range(N)]
    # Consider a vapor fraction of more than 0.5 a vapor
#    return flash_inner_loop(zs, Ks)[0] - 0.5
    # Go back to the error once unit tested
    return Rachford_Rice_flash_error(V_over_F=0.5, zs=zs, Ks=Ks)


def vapor_score_Poling(kappa):
    r'''Compute a vapor score representing how vapor-like a phase is
    (higher, above zero = more vapor like) using the isothermal compressibility
    `kappa` concept by Poling [1]_.

    .. math::
        \text{score} = (\kappa - 0.005 \text{atm}^{-1})

    Parameters
    ----------
    kappa : float
        Isothermal coefficient of compressibility, [1/Pa]

    Returns
    -------
    score : float
        Vapor like score, [-]

    Examples
    --------
    CO2 vapor properties computed with Peng-Robinson at 300 K and 1 bar:

    >>> vapor_score_Poling(1.0054239121594122e-05)
    1.013745778995

    n-hexane liquid properties computed with Peng-Robinson at 300 K and 10 bar:

    >>> vapor_score_Poling(2.121777078782957e-09)
    -0.00478501093

    Notes
    -----
    A second criteria which is not implemented as it does not fit with the
    scoring concept is for liquids:

    .. math::
        \frac{0.9}{P} < \beta < \frac{3}{P}

    References
    ----------
    .. [1] Poling, Bruce E., Edward A. Grens, and John M. Prausnitz.
       "Thermodynamic Properties from a Cubic Equation of State: Avoiding
       Trivial Roots and Spurious Derivatives." Industrial & Engineering
       Chemistry Process Design and Development 20, no. 1 (January 1, 1981):
       127-30. https://doi.org/10.1021/i200012a019.
    '''
    return kappa*101325 - .005

def vapor_score_PIP(V, dP_dT, dP_dV, d2P_dV2, d2P_dVdT):
    r'''Compute a vapor score representing how vapor-like a phase is
    (higher, above zero = more vapor like) using the PIP concept.

    .. math::
        \text{score} = -(\Pi - 1)

    .. math::
        \Pi = V \left[\frac{\frac{\partial^2 P}{\partial V \partial T}}
        {\frac{\partial P }{\partial T}}- \frac{\frac{\partial^2 P}{\partial
        V^2}}{\frac{\partial P}{\partial V}} \right]

    Parameters
    ----------
    V : float
        Molar volume at `T` and `P`, [m^3/mol]
    dP_dT : float
        Derivative of `P` with respect to `T`, [Pa/K]
    dP_dV : float
        Derivative of `P` with respect to `V`, [Pa*mol/m^3]
    d2P_dV2 : float
        Second derivative of `P` with respect to `V`, [Pa*mol^2/m^6]
    d2P_dVdT : float
        Second derivative of `P` with respect to both `V` and `T`, [Pa*mol/m^3/K]

    Returns
    -------
    score : float
        Vapor like score, [-]

    Examples
    --------
    CO2 vapor properties computed with Peng-Robinson at 300 K and 1 bar:

    >>> vapor_score_PIP(0.024809176851423774, 337.0119286073647, -4009021.959558917, 321440573.3615088, -13659.63987996052)
    0.016373735005

    n-hexane liquid properties computed with Peng-Robinson at 300 K and 10 bar:

    >>> vapor_score_PIP(0.00013038156684574785, 578477.8796379718, -3614798144591.8984, 4.394997991022487e+17, -20247865009.795322)
    -10.288635225

    References
    ----------
    .. [1] Venkatarathnam, G., and L. R. Oellrich. "Identification of the Phase
       of a Fluid Using Partial Derivatives of Pressure, Volume, and
       Temperature without Reference to Saturation Properties: Applications in
       Phase Equilibria Calculations." Fluid Phase Equilibria 301, no. 2
       (February 25, 2011): 225-33. doi:10.1016/j.fluid.2010.12.001.
    '''
    return -(phase_identification_parameter(V, dP_dT, dP_dV, d2P_dV2, d2P_dVdT) - 1.0)

def vapor_score_Bennett_Schmidt(dbeta_dT):
    r'''Compute a vapor score representing how vapor-like a phase is
    (higher, above zero = more vapor like) using the Bennet-Schmidt
    temperature derivative of isobaric expansion suggestion.

    .. math::
        \text{score} = -\left(\frac{\partial \beta}{\partial T}\right)

    Parameters
    ----------
    dbeta_dT : float
        Temperature derivative of isobaric coefficient of a thermal
        expansion, [1/K^2]

    Returns
    -------
    score : float
        Vapor like score, [-]

    Examples
    --------
    CO2 vapor properties computed with Peng-Robinson at 300 K and 1 bar:

    >>> vapor_score_Bennett_Schmidt(-1.1776172267959163e-05)
    1.1776172267959163e-05

    n-hexane liquid properties computed with Peng-Robinson at 300 K and 10 bar:

    >>> vapor_score_Bennett_Schmidt(7.558572848883679e-06)
    -7.558572848883679e-06

    References
    ----------
    .. [1] Bennett, Jim, and Kurt A. G. Schmidt. "Comparison of Phase
       Identification Methods Used in Oil Industry Flow Simulations." Energy &
       Fuels 31, no. 4 (April 20, 2017): 3370-79.
       https://doi.org/10.1021/acs.energyfuels.6b02316.
    '''
    return -dbeta_dT

def vapor_score_traces(zs, CASs, Tcs, trace_CASs=['74-82-8', '7727-37-9'],
                       min_trace=0.0):
    r'''Compute a vapor score representing how vapor-like a phase is
    (higher, above zero = more vapor like) using the concept of which phase has
    the most of the lightest compound. This nicely sidesteps issues in many
    other methods, at the expense that it cannot be applied when there is only
    one phase and it is not smart enough to handle liquid-liquid cases.

    If no trace components are present, the component with the lowest critical
    temperature's concentration is returned. Because of the way this is
    implemented, the score is always larger than 1.0.


    Parameters
    ----------
    zs : list[float]
        Mole fractions of the phase being identified, [-]
    CASs : list[str]
        CAS numbers of all components, [-]
    Tcs : list[float]
        Critical temperatures of all species, [K]
    trace_CASs : list[str]
        Trace components to use for identification; if more than one component
        is given, the first component present in both `CASs` and `trace_CASs`
        is the one used, [-]
    min_trace : float
        Minimum concentration to make a phase appear vapor-like; subtracted
        from the concentration which would otherwise be returned, [-]

    Returns
    -------
    score : float
        Vapor like score, [-]

    Examples
    --------
    A flash of equimolar CO2/n-hexane at 300 K and 1 MPa is computed, and there
    is a two phase solution. The phase must be identified for each result:

    Liquid-like phase:

    >>> vapor_score_traces(zs=[.218, .782], Tcs=[304.2, 507.6], CASs=['124-38-9', '110-54-3'])
    0.218

    Vapor-like phase:

    >>> vapor_score_traces(zs=[.975, .025], Tcs=[304.2, 507.6], CASs=['124-38-9', '110-54-3'])
    0.975

    Notes
    -----
    '''
    # traces should be the lightest species - high = more vapor like
    if trace_CASs is not None:
        for trace_CAS in trace_CASs:
            try:
                return zs[CASs.index(trace_CAS)] - min_trace
            except ValueError:
                # trace component not in mixture
                pass

    # Return the composition of the compound with the lowest critical temp
    comp = 0.0
    Tc_min = 1e100
    for i in range(len(zs)):
        if Tcs[i] < Tc_min:
            comp = zs[i]
            Tc_min = Tcs[i]
    return comp - min_trace


VL_ID_TPC = 'Tpc'
VL_ID_VPC = 'Vpc'
VL_ID_TPC_VC_WEIGHTED = 'Tpc Vpc weighted'
VL_ID_TPC_VPC = 'Tpc Vpc'
VL_ID_WILSON = 'Wilson'
VL_ID_POLING = 'Poling'
VL_ID_PIP = 'PIP'
VL_ID_BS = 'Bennett-Schmidt'
VL_ID_TRACES = 'Traces'

VL_ID_METHODS = [VL_ID_TPC, VL_ID_VPC, VL_ID_TPC_VC_WEIGHTED, VL_ID_TPC_VPC,
                 VL_ID_WILSON, VL_ID_POLING, VL_ID_PIP, VL_ID_BS, VL_ID_TRACES]
'''List of all the methods available to perform the Vapor-Liquid phase ID.
'''
S_ID_D2P_DVDT = 'd2P_dVdT'
S_ID_METHODS = [S_ID_D2P_DVDT]
'''List of all the methods available to perform the solid-liquid phase ID.
'''

def score_phases_S(phases, constants, correlations, method=S_ID_D2P_DVDT,
                   S_ID_settings=None):
    r'''Score all phases according to how wolid they appear given the
    provided parameters and a selected method.

    A score above zero indicates a solid. More than one phase may have
    a score above zero. A score under zero means the phase is a liquid or gas.

    Parameters
    ----------
    phases : list[:obj:`thermo.phases.Phase`]
        Phases to be identified and sorted, [-]
    constants : :obj:`ChemicalConstantsPackage <thermo.chemical_package.ChemicalConstantsPackage>`
        Constants used in the identification, [-]
    correlations : :obj:`PropertyCorrelationsPackage <thermo.chemical_package.PropertyCorrelationsPackage>`
        Correlations used in the identification, [-]
    method : str
        Setting configuring how the scoring is performed; one of
        ('d2P_dVdT',), [-]
    S_ID_settings : dict[str : float] or None, optional
        Additional configuration options for solid-liquid phase ID, [-]

    Returns
    -------
    scores : list[float]
        Scores for the phases in the order provided, [-]
    '''
    # The higher the score (above zero), the more solid-like
    if method == S_ID_D2P_DVDT:
        scores = [i.d2P_dVdT() for i in phases]
    return scores

def score_phases_VL(phases, constants, correlations, method):
    r'''Score all phases given the provided parameters and a selected method.

    A score above zero indicates a potential gas. More than one phase may have
    a score above zero, in which case the highest scoring phase is the gas,
    and the other is a liquid.

    Parameters
    ----------
    phases : list[:obj:`thermo.phases.Phase`]
        Phases to be identified and sorted, [-]
    constants : :obj:`ChemicalConstantsPackage <thermo.chemical_package.ChemicalConstantsPackage>`
        Constants used in the identification, [-]
    correlations : :obj:`PropertyCorrelationsPackage <thermo.chemical_package.PropertyCorrelationsPackage>`
        Correlations used in the identification, [-]
    method : str
        Setting configuring how the scoring is performed; one of
        'Tpc', 'Vpc', 'Tpc Vpc weighted', 'Tpc Vpc', 'Wilson', 'Poling',
        'PIP', 'Bennett-Schmidt', 'Traces', [-]

    Returns
    -------
    scores : list[float]
        Scores for the phases in the order provided, [-]

    Notes
    -----

    Examples
    --------
    >>> from thermo import ChemicalConstantsPackage, PropertyCorrelationsPackage, CEOSGas, CEOSLiquid, PRMIX, HeatCapacityGas
    >>> constants = ChemicalConstantsPackage(CASs=['124-38-9', '110-54-3'], Vcs=[9.4e-05, 0.000368], MWs=[44.0095, 86.17536], names=['carbon dioxide', 'hexane'], omegas=[0.2252, 0.2975], Pcs=[7376460.0, 3025000.0], Tbs=[194.67, 341.87], Tcs=[304.2, 507.6], Tms=[216.65, 178.075])
    >>> correlations = PropertyCorrelationsPackage(constants=constants, skip_missing=True, HeatCapacityGases=[HeatCapacityGas(poly_fit=(50.0, 1000.0, [-3.1115474168865828e-21, 1.39156078498805e-17, -2.5430881416264243e-14, 2.4175307893014295e-11, -1.2437314771044867e-08, 3.1251954264658904e-06, -0.00021220221928610925, 0.000884685506352987, 29.266811602924644])), HeatCapacityGas(poly_fit=(200.0, 1000.0, [1.3740654453881647e-21, -8.344496203280677e-18, 2.2354782954548568e-14, -3.4659555330048226e-11, 3.410703030634579e-08, -2.1693611029230923e-05, 0.008373280796376588, -1.356180511425385, 175.67091124888998]))])
    >>> T, P, zs = 300.0, 1e6, [.5, .5]
    >>> eos_kwargs = {'Pcs': constants.Pcs, 'Tcs': constants.Tcs, 'omegas': constants.omegas}
    >>> gas = CEOSGas(PRMIX, eos_kwargs, HeatCapacityGases=correlations.HeatCapacityGases, T=T, P=P, zs=zs)
    >>> liq = CEOSLiquid(PRMIX, eos_kwargs, HeatCapacityGases=correlations.HeatCapacityGases, T=T, P=P, zs=zs)

    A sampling of different phase identification methods is below:

    >>> score_phases_VL([gas, liq], constants, correlations, method='PIP')
    [1.6409446310, -7.5692120928]
    >>> score_phases_VL([gas, liq], constants, correlations, method='Vpc')
    [0.00144944049, -0.0001393075288]
    >>> score_phases_VL([gas, liq], constants, correlations, method='Tpc Vpc')
    [113.181283525, -29.806038704]
    >>> score_phases_VL([gas, liq], constants, correlations, method='Bennett-Schmidt')
    [0.0003538299416, -2.72255439503e-05]
    >>> score_phases_VL([gas, liq], constants, correlations, method='Poling')
    [0.1767828268, -0.004516837897]
    '''
    # The higher the score (above zero), the more vapor-like
    if phases:
        T = phases[0].T
    if method == VL_ID_TPC:
        Tcs = constants.Tcs
        scores = [vapor_score_Tpc(T, Tcs, i.zs) for i in phases]
    elif method == VL_ID_VPC:
        Vcs = constants.Vcs
        scores = [vapor_score_Vpc(i.V(), Vcs, i.zs) for i in phases]
    elif method == VL_ID_TPC_VC_WEIGHTED:
        Tcs = constants.Tcs
        Vcs = constants.Vcs
        scores = [vapor_score_Tpc_weighted(T, Tcs, Vcs, i.zs) for i in phases]
    elif method == VL_ID_TPC_VPC:
        Tcs = constants.Tcs
        Vcs = constants.Vcs
        scores = [vapor_score_Tpc_Vpc(T, i.V(), Tcs, Vcs, i.zs) for i in phases]
    elif method == VL_ID_WILSON:
        Tcs = constants.Tcs
        Pcs = constants.Pcs
        omegas = constants.omegas
        scores = [vapor_score_Wilson(T, i.P, i.zs, Tcs, Pcs, omegas) for i in phases]
    elif method == VL_ID_POLING:
        scores = [vapor_score_Poling(i.kappa()) for i in phases]
    elif method == VL_ID_PIP:
        scores = [-(i.PIP() - 1.00000000000001) for i in phases]
#        scores = [vapor_score_PIP(i.V(), i.dP_dT(), i.dP_dV(),
#                                           i.d2P_dV2(), i.d2P_dVdT()) for i in phases]
    elif method == VL_ID_BS:
        scores = [vapor_score_Bennett_Schmidt(i.disobaric_expansion_dT()) for i in phases]
    elif method == VL_ID_TRACES:
        CASs = constants.CASs
        Tcs = constants.Tcs
        scores = [vapor_score_traces(i.zs, CASs, Tcs=Tcs) for i in phases]
    return scores


def identity_phase_states(phases, constants, correlations, VL_method=VL_ID_PIP,
                          S_method=S_ID_D2P_DVDT,
                          VL_ID_settings=None, S_ID_settings=None,
                          skip_solids=False):
    r'''Identify and the actial phase of all the given phases given the
    provided settings.

    Parameters
    ----------
    phases : list[:obj:`Phase <thermo.phases.Phase>`]
        Phases to be identified and sorted, [-]
    constants : :obj:`ChemicalConstantsPackage <thermo.chemical_package.ChemicalConstantsPackage>`
        Constants used in the identification, [-]
    correlations : :obj:`PropertyCorrelationsPackage <thermo.chemical_package.PropertyCorrelationsPackage>`
        Correlations used in the identification, [-]
    VL_method : str, optional
        One of :obj:`VL_ID_METHODS`, [-]
    S_method : str, optional
        One of :obj:`S_ID_METHODS`, [-]
    VL_ID_settings : dict[str : float] or None, optional
        Additional configuration options for vapor-liquid phase ID, [-]
    S_ID_settings : dict[str : float] or None, optional
        Additional configuration options for solid-liquid phase ID, [-]
    skip_solids : bool
        Set this to True if no phases are provided which can represent a solid phase, [-]

    Returns
    -------
    gas : :obj:`Phase <thermo.phases.Phase>`
        Gas phase, if one was identified, [-]
    liquids : list[:obj:`Phase <thermo.phases.Phase>`]
        Liquids that were identified and sorted, [-]
    solids : list[:obj:`Phase <thermo.phases.Phase>`]
        Solids that were identified and sorted, [-]
    '''
    # TODO - optimize
    # Takes a while

    force_phases = [i.force_phase for i in phases]
    forced = True
    phases_to_ID = []
    phases_to_ID_idxs = []
    for i, s in enumerate(force_phases):
        if s is None:
            forced = False
            # TODO avoid scoring phases with force phase
            phases_to_ID.append(phases[i])
            phases_to_ID_idxs.append(i)

    if not forced:
        VL_scores = score_phases_VL(phases, constants, correlations,
                                    method=VL_method)
        if not skip_solids:
            S_scores = score_phases_S(phases, constants, correlations,
                                      method=S_method, S_ID_settings=S_ID_settings)

    solids = []
    liquids = []
    possible_gases = []
    possible_gas_scores = []

    for i in range(len(phases)):
        if force_phases[i] is not None:
            if force_phases[i] == 'l':
                liquids.append(phases[i])
            if force_phases[i] == 's':
                solids.append(phases[i])
            if force_phases[i] == 'g':
                possible_gases.append(phases[i])
        elif not skip_solids and S_scores[i] >= 0.0:
            solids.append(phases[i])
        elif VL_scores[i] >= 0.0:
            possible_gases.append(phases[i])
            possible_gas_scores.append(VL_scores[i])
        else:
            liquids.append(phases[i])

    # Handle multiple matches as gas
    possible_gas_count = len(possible_gases)
    if possible_gas_count > 1:
        gas = possible_gases[possible_gas_scores.index(max(possible_gas_scores))]
        for possible_gas in possible_gases:
            if possible_gas is not gas:
                liquids.append(possible_gas)
        possible_gases[:] = (gas,)
    elif possible_gas_count == 1:
        gas = possible_gases[0]
    else:
        gas = None

    return gas, liquids, solids



DENSITY_MASS = 'DENSITY_MASS'
DENSITY = 'DENSITY'
ISOTHERMAL_COMPRESSIBILITY = 'ISOTHERMAL_COMPRESSIBILITY'
HEAT_CAPACITY = 'HEAT_CAPACITY'
L_SORT_PROPS = S_SORT_PROPS = [DENSITY_MASS, DENSITY, ISOTHERMAL_COMPRESSIBILITY,
                HEAT_CAPACITY]

WATER_FIRST = 'water first'
WATER_LAST = 'water last'
WATER_NOT_SPECIAL = 'water not special'

WATER_SORT_METHODS = [WATER_FIRST, WATER_LAST, WATER_NOT_SPECIAL]

KEY_COMPONENTS_SORT = 'key components'
PROP_SORT = 'prop'
SOLID_SORT_METHODS = LIQUID_SORT_METHODS = [PROP_SORT, KEY_COMPONENTS_SORT]

def key_cmp_sort(phases, cmps, cmps_neg):
    # TODO
    raise NotImplementedError
#    return phases

def mini_sort_phases(phases, sort_method, prop, cmps, cmps_neg,
                     reverse=True, constants=None):
    if sort_method == PROP_SORT:
        if prop == DENSITY_MASS:
            keys = []
            MWs = constants.MWs
            for p in phases:
                zs = p.zs
                MW = 0.0
                for i in constants.cmps:
                    MW += zs[i]*MWs[i]
                keys.append(Vm_to_rho(p.V(), MW))

            # for i in phases:
            #     i.constants = constants
            # keys = [i.rho_mass() for i in phases]
        elif prop == DENSITY:
            keys = [i.rho() for i in phases]
        elif prop == ISOTHERMAL_COMPRESSIBILITY:
            keys = [i.isobaric_expansion() for i in phases]
        elif prop == HEAT_CAPACITY:
            keys = [i.Cp() for i in phases]
        phases = [p for _, p in sorted(zip(keys, phases))]
        if reverse:
            phases.reverse()
    elif sort_method == KEY_COMPONENTS_SORT:
        phases = key_cmp_sort(phases, cmps, cmps_neg)
        if not reverse:
            phases.reverse()
    return phases

def sort_phases(liquids, solids, constants, settings):
    r'''Identify and sort all phases given the provided parameters.
    This is not a thermodynamic concept; it is just a convinience
    method to make the results of the flash more consistent, because
    the flash algorithms don't care about density or ordering
    the phases.

    Parameters
    ----------
    liquids : list[:obj:`Phase <thermo.phases.Phase>`]
        Liquids that were identified, [-]
    solids : list[:obj:`Phase <thermo.phases.Phase>`]
        Solids that were identified, [-]
    constants : :obj:`ChemicalConstantsPackage <thermo.chemical_package.ChemicalConstantsPackage>`
        Constants used in the identification, [-]
    correlations : :obj:`PropertyCorrelationsPackage <thermo.chemical_package.PropertyCorrelationsPackage>`
        Correlations used in the identification, [-]
    settings : :obj:`BulkSettings <thermo.bulk.BulkSettings>`
        Settings object controlling the phase sorting, [-]

    Returns
    -------
    liquids : list[:obj:`Phase <thermo.phases.Phase>`]
        Liquids that were identified and sorted, [-]
    solids : list[:obj:`Phase <thermo.phases.Phase>`]
        Solids that were identified and sorted, [-]

    Notes
    -----
    The settings object uses the preferences
    `liquid_sort_method`, `liquid_sort_prop`,
    `liquid_sort_cmps`, `liquid_sort_cmps_neg`,
    and `phase_sort_higher_first`.

    Examples
    --------

    '''

    if len(liquids) > 1:
        liquids = mini_sort_phases(liquids, sort_method=settings.liquid_sort_method,
                         prop=settings.liquid_sort_prop,
                         cmps=settings.liquid_sort_cmps,
                         cmps_neg=settings.liquid_sort_cmps_neg,
                         reverse=settings.phase_sort_higher_first, constants=constants)

        # Handle water special
        if settings.water_sort != WATER_NOT_SPECIAL:
            # water phase - phase with highest fraction water
            water_index = constants.water_index
            if water_index is not None:
                water_zs = [i.zs[water_index] for i in liquids]
                water_max_zs = max(water_zs)
                if water_max_zs > 1e-4:
                    water_phase_index = water_zs.index(water_max_zs)
                    water = liquids.pop(water_phase_index)
                    if settings.water_sort == WATER_LAST:
                        liquids.append(water)
                    elif settings.water_sort == WATER_FIRST:
                        liquids.insert(0, water)
    if len(solids) > 1:
        solids = mini_sort_phases(solids, sort_method=settings.solid_sort_method,
                         prop=setings.solid_sort_prop,
                         cmps=settings.solid_sort_cmps,
                         cmps_neg=settings.solid_sort_cmps_neg,
                         reverse=settings.phase_sort_higher_first, constants=constants)
    return liquids, solids


def identify_sort_phases(phases, betas, constants, correlations, settings,
                         skip_solids=False):
    r'''Identify and sort all phases given the provided parameters.

    Parameters
    ----------
    phases : list[:obj:`Phase <thermo.phases.Phase>`]
        Phases to be identified and sorted, [-]
    betas : list[float]
        Phase molar fractions, [-]
    constants : :obj:`ChemicalConstantsPackage <thermo.chemical_package.ChemicalConstantsPackage>`
        Constants used in the identification, [-]
    correlations : :obj:`PropertyCorrelationsPackage <thermo.chemical_package.PropertyCorrelationsPackage>`
        Correlations used in the identification, [-]
    settings : :obj:`BulkSettings <thermo.bulk.BulkSettings>`
        Settings object controlling the phase ID, [-]
    skip_solids : bool
        Set this to True if no phases are provided which can represent a solid phase, [-]

    Returns
    -------
    gas : :obj:`Phase <thermo.phases.Phase>`
        Gas phase, if one was identified, [-]
    liquids : list[:obj:`Phase <thermo.phases.Phase>`]
        Liquids that were identified and sorted, [-]
    solids : list[:obj:`Phase <thermo.phases.Phase>`]
        Solids that were identified and sorted, [-]
    betas : list[float]
        Sorted phase molar fractions, in order (gas, liquids..., solids...) [-]

    Notes
    -----
    This step is very important as although phase objects are
    designed to represent a single phase, cubic equations of state
    can be switched back and forth by the flash algorithms.
    Thermodynamics doesn't care about gases, liquids, or solids;
    it just cares about minimizing Gibbs energy!

    Examples
    --------
    A butanol-water-ethanol flash yields three phases. For brevity we skip
    the flash and initialize our `gas`, `liq0`, and `liq1` object with the
    correct phase composition. Then we identify the phases into liquid,
    gas, and solid.

    >>> from thermo import ChemicalConstantsPackage, PropertyCorrelationsPackage, HeatCapacityGas, SRKMIX, CEOSGas, CEOSLiquid
    >>> constants = ChemicalConstantsPackage(Tcs=[563.0, 647.14, 514.0], Vcs=[0.000274, 5.6e-05, 0.000168], Pcs=[4414000.0, 22048320.0, 6137000.0], omegas=[0.59, 0.344, 0.635], MWs=[74.1216, 18.01528, 46.06844], CASs=['71-36-3', '7732-18-5', '64-17-5'])
    >>> properties = PropertyCorrelationsPackage(constants=constants, skip_missing=True,
    ...                                     HeatCapacityGases=[HeatCapacityGas(load_data=False, poly_fit=(50.0, 1000.0, [-3.787200194613107e-20, 1.7692887427654656e-16, -3.445247207129205e-13, 3.612771874320634e-10, -2.1953250181084466e-07, 7.707135849197655e-05, -0.014658388538054169, 1.5642629364740657, -7.614560475001724])),
    ...                                     HeatCapacityGas(load_data=False, poly_fit=(50.0, 1000.0, [5.543665000518528e-22, -2.403756749600872e-18, 4.2166477594350336e-15, -3.7965208514613565e-12, 1.823547122838406e-09, -4.3747690853614695e-07, 5.437938301211039e-05, -0.003220061088723078, 33.32731489750759])),
    ...                                     HeatCapacityGas(load_data=False, poly_fit=(50.0, 1000.0, [-1.162767978165682e-20, 5.4975285700787494e-17, -1.0861242757337942e-13, 1.1582703354362728e-10, -7.160627710867427e-08, 2.5392014654765875e-05, -0.004732593693568646, 0.5072291035198603, 20.037826650765965])),], )
    >>> eos_kwargs = dict(Tcs=constants.Tcs, Pcs=constants.Pcs, omegas=constants.omegas)
    >>> gas = CEOSGas(SRKMIX, eos_kwargs, HeatCapacityGases=properties.HeatCapacityGases)
    >>> liq = CEOSLiquid(SRKMIX, eos_kwargs, HeatCapacityGases=properties.HeatCapacityGases)
    >>> T, P = 361, 1e5
    >>> gas = gas.to(T=T, P=P, zs=[0.2384009970908655, 0.5786839935180925, 0.1829150093910419])
    >>> liq0 = liq.to(T=T, P=P, zs=[7.619975052238032e-05, 0.9989622883894993, 0.0009615118599781474])
    >>> liq1 = liq.to(T=T, P=P, zs=[0.6793120076703771, 0.19699746328631124, 0.12369052904331178])
    >>> res = identity_phase_states(phases=[liq0, liq1, gas], constants=constants, correlations=properties, VL_method='PIP')
    >>> res[0] is gas, res[1][0] is liq0, res[1][1] is liq1, res[2]
    (True, True, True, [])
    '''
    gas, liquids, solids = identity_phase_states(phases, constants, correlations,
                              VL_method=settings.VL_ID,
                              S_method=settings.S_ID,
                              VL_ID_settings=settings.VL_ID_settings,
                              S_ID_settings=settings.S_ID_settings,
                              skip_solids=skip_solids)
    if len(betas) == 1:
        return gas, liquids, solids, betas
    if liquids or solids:
        liquids, solids = sort_phases(liquids, solids, constants, settings)
    if betas is not None:
        # Probably worth adding a fast path here which avoids index
        new_betas = []
        if gas is not None:
            # Cannot use .index on the betas lsit as the objects define __hash__.
            for i, p in enumerate(phases):
                if p is gas:
                    new_betas.append(betas[i])
                    break
        for liquid in liquids:
            for i, p in enumerate(phases):
                if p is liquid:
                    new_betas.append(betas[i])
                    break
        for solid in solids:
            for i, p in enumerate(phases):
                if p is solid:
                    new_betas.append(betas[i])
                    break
        betas = new_betas
    return gas, liquids, solids, betas





# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2021 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This module contains functionality for serializing classes.

For reporting bugs, adding feature requests, or submitting pull requests,
please use the `GitHub issue tracker <https://github.com/CalebBell/chemicals/>`_.

.. contents:: :local:
'''
# This module SHOULD NOT import anything from thermo
from fluids.numerics import numpy as np
from chemicals.utils import PY37

__all__ = ['object_from_json', 'json_default']
try:
    array = np.array
    int_types = frozenset([np.short, np.ushort, np.intc, np.uintc, np.int_,
                           np.uint, np.longlong, np.ulonglong])
    float_types = frozenset([np.float16, np.single, np.double, np.longdouble,
                             np.csingle, np.cdouble, np.clongdouble])
    ndarray = np.ndarray

except:
    pass


BasicNumpyEncoder = None
def build_numpy_encoder():
    '''Create a basic numpy encoder for json applications. All np ints become
    Python ints; numpy floats become python floats; and all arrays become
    lists-of-lists of floats, ints, bools, or complexes according to Python's
    rules.
    '''
    global BasicNumpyEncoder, json, int_types, float_types

    import json
    JSONEncoder = json.JSONEncoder

    class BasicNumpyEncoder(JSONEncoder):
        def default(self, obj):
            t = type(obj)
            if t is ndarray:
                return obj.tolist()
            elif t in int_types:
                return int(obj)
            elif t in float_types:
                return float(obj)
            return JSONEncoder.default(self, obj)

def arrays_to_lists(obj):
    t = type(obj)
    if t is dict:
        # Do not modify objects in place
        obj = obj.copy()
        for k, v in obj.items():
            obj[k] = arrays_to_lists(v)
    elif t is tuple:
        return tuple(arrays_to_lists(v) for v in obj)
    elif t is set:
        # Can't put arrays in sets
        return obj
#        return set(arrays_to_lists(v) for v in obj)
    elif t is ndarray:
        return obj.tolist()
    elif t in int_types:
        return int(obj)
    elif t in float_types:
        return float(obj)
    return obj


def naive_lists_to_arrays(obj):
    t = type(obj)
    if t is dict:
        # Do not modify objects in place
        obj = obj.copy()
        for k, v in obj.items():
            obj[k] = naive_lists_to_arrays(v)
    elif t is tuple:
        return tuple(naive_lists_to_arrays(v) for v in obj)
    elif t is set:
        return set(naive_lists_to_arrays(v) for v in obj)
    elif t is list:
        if len(obj) >= 2 and type(obj[0]) is list:
            # Handle tuples of different sized arrays
            try:
                length = len(obj[0])
                for v in obj[1:]:
                    if len(v) != length:
                        return tuple(array(v) for v in obj)
            except:
                pass
        return array(obj)
    return obj

BasicNumpyDecoder = None
def build_numpy_decoder():
    global BasicNumpyDecoder, json
    import json

    class BasicNumpyDecoder(json.JSONDecoder):
        def __init__(self, *args, **kwargs):
            json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

        def object_hook(self, obj):
            return naive_lists_to_arrays(obj)


def _load_orjson():
    global orjson
    import orjson

def dump_json_np(obj, library='json'):
    '''Serialization tool that handles numpy arrays. By default this will
    use the standard library json, but this can also use orjson.'''
    if library == 'json':
        if BasicNumpyEncoder is None:
            build_numpy_encoder()
        return json.dumps(obj, cls=BasicNumpyEncoder)
    elif library == 'orjson':
        if orjson is None:
            _load_orjson()
        opt = orjson.OPT_SERIALIZE_NUMPY
        return orjson.dumps(obj, option=opt)

def load_json_np(obj, library='json'):
    '''Serialization tool that handles numpy arrays. By default this will
    use the standard library json, but this can also use orjson.'''
    if library == 'json':
        if BasicNumpyDecoder is None:
            build_numpy_decoder()
        return json.loads(obj, cls=BasicNumpyDecoder)
    elif library == 'orjson':
        if orjson is None:
            _load_orjson()
        opt = orjson.OPT_SERIALIZE_NUMPY
        return orjson.loads(obj, option=opt)

json_loaded = False
def _load_json():
    global json
    import json
    json_loaded = True

global json, orjson
orjson = None

if PY37:
    def __getattr__(name):
        global json, json_loaded
        if name == 'json':
            import json
            json_loaded = True
            return json
        raise AttributeError("module %s has no attribute %s" %(__name__, name))
else:
    import json
    json_loaded = True





def object_from_json(json_object):
    if 'py/object' in json_object:
        pth = json_object['py/object']
        # TODO: Cache these lookups
        bits = pth.split('.')
        mod = '.'.join(bits[:-1])
        cls_name = bits[-1]
        obj = getattr(sys.modules[mod], cls_name)

        return obj.from_json(json_object)
    raise ValueError("Could not recognize object")


def json_default(obj):
    if hasattr(obj, 'as_json'):
        return obj.as_json()
    raise TypeError()
